package jp.sourceforge.pdt_tools.adjustencoding;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;

import jp.sourceforge.pdt_tools.adjustencoding.filters.IFilter;
import jp.sourceforge.pdt_tools.adjustencoding.filters.SampleFilter;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.content.IContentDescription;
import org.eclipse.core.runtime.content.IContentType;
import org.eclipse.core.runtime.content.IContentTypeManager;
import org.eclipse.ui.IStartup;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.mozilla.universalchardet.UniversalDetector;
import org.osgi.framework.BundleContext;

public class AdjustEncoding extends AbstractUIPlugin implements IStartup {

	// The plug-in ID
	public static final String PLUGIN_ID = "jp.sourceforge.pdt_tools.adjustencoding"; //$NON-NLS-1$

	// The shared instance
	private static AdjustEncoding plugin;

	// ContentType Text
	private static final IContentType CT_TEXT = Platform
			.getContentTypeManager()
			.getContentType(IContentTypeManager.CT_TEXT);

	// detects encoding within first 4K bytes
	private int BUFFER_SIZE = 4096;

	// inclusion/exclusion filter
	private IFilter filter;

	public AdjustEncoding() {
		filter = new SampleFilter();
	}

	public void start(BundleContext context) throws Exception {
		super.start(context);
		plugin = this;
	}

	public void stop(BundleContext context) throws Exception {
		plugin = null;
		super.stop(context);
	}

	/**
	 * Returns the shared instance
	 * 
	 * @return the shared instance
	 */
	public static AdjustEncoding getDefault() {
		return plugin;
	}

	public void log(Throwable e) {
		getDefault().getLog().log(
				new Status(IStatus.ERROR, PLUGIN_ID, e.getLocalizedMessage()));
	}

	public void log(int severity, String msg) {
		getDefault().getLog().log(new Status(severity, PLUGIN_ID, msg));
	}

	@Override
	public void earlyStartup() {
		ResourcesPlugin.getWorkspace().addResourceChangeListener(
				new ResourceChangeListener());
	}

	/**
	 * Listening resources(text files) added
	 */
	private class ResourceChangeListener implements IResourceChangeListener {

		@Override
		public void resourceChanged(IResourceChangeEvent event) {
			if (event.getType() == IResourceChangeEvent.POST_CHANGE) {
				processDelta(event.getDelta());
			}
		}

		private void processDelta(IResourceDelta delta) {
			switch (delta.getKind()) {
			case IResourceDelta.ADDED:
			case IResourceDelta.CHANGED:
				break;
			default:
				return;
			}
			for (IResourceDelta child : delta.getAffectedChildren()) {
				IResource resource = child.getResource();
				if (resource.getType() == IResource.FOLDER) {
					if (filter.omit(resource)) {
						continue;
					}
				}
				processDelta(child);
			}
			if (delta.getKind() == IResourceDelta.ADDED) {
				IResource resource = delta.getResource();
				if (resource instanceof IFile) {
					try {
						if (shouldAdjust((IFile) resource)) {
							adjustEncoding((IFile) resource);
						}
					} catch (CoreException e) {
						log(e);
					}
				}
			}
		}
	}

	private boolean shouldAdjust(IFile file) throws CoreException {
		if (file.getCharset(false) != null) {
			return false;
		}
		if (filter.omit(file)) {
			return false;
		}
		IContentDescription description = file.getContentDescription();
		if (description == null) {
			// unknown(unhandled) file type => assumed text file
			return true;
		}
		IContentType type = description.getContentType();
		if (type != null && type.isKindOf(CT_TEXT)) {
			return true;
		}
		return false;
	}

	public void adjustEncoding(final IFile file) throws CoreException {
		byte[] buf = new byte[BUFFER_SIZE];
		int size = 0;
		InputStream contents = file.getContents();
		try {
			size = contents.read(buf, 0, buf.length);
		} catch (IOException e) {
			log(e);
		} finally {
			try {
				contents.close();
			} catch (IOException e) {
			}
		}
		if (size == 0) {
			return;
		}
		UniversalDetector detector = new UniversalDetector(null);
		detector.handleData(buf, 0, size);
		detector.dataEnd();
		String detectedName = detector.getDetectedCharset();
		if (detectedName == null) {
			if (Platform.inDebugMode()) {
				String msg = String.format(
						"Could not detect encoding : %s", file.getLocation()); //$NON-NLS-1$
				log(IStatus.WARNING, msg);
			}
			return;
		}
		final Charset detected = Charset.forName(detectedName);
		Charset defined = Charset.forName(file.getCharset(true));
		if (!detected.equals(defined)) {
			WorkspaceJob job = new WorkspaceJob("") { //$NON-NLS-1$
				@Override
				public IStatus runInWorkspace(IProgressMonitor monitor)
						throws CoreException {
					file.setCharset(detected.name(), new NullProgressMonitor());
					return Status.OK_STATUS;
				}
			};
			job.schedule();
			if (Platform.inDebugMode()) {
				String msg = String.format("Sets %s instead of %s : %s", //$NON-NLS-1$
						detected.name(), defined.name(), file.getLocation());
				log(IStatus.INFO, msg);
			}
		}
	}
}
